from .__version__ import (
    __title__,
    __url__,
    __version__,
    __author__,
    __author_email__,
    __license__,
    __copyright__,
)

from .auto_switch_providers import AutoSwitchProviders
from .http_service import HttpService
from .utils.helpers import merge_child, nested_dict_get



from json import dumps
from time import perf_counter

from .services.redis_service import RedisClient


class CacheService:
    def __init__(self, config: dict = None) -> None:
        self.client = None
        if config is not None:
            self.config = config
            start_time = perf_counter()
            self.client = RedisClient(self.config).get_client()
            if self.client:
                try:
                    self.client.ping()
                    self.status = True
                    exec_time = perf_counter() - start_time
                    host_name = self.config.get("host")
                    print(
                        f"⚡️ Redis enabled! (ping: {exec_time*1000:.2f} ms, host: {host_name})"
                    )
                except Exception:
                    self.status = False
                    print("⛔️ Redis disabled!")
                    print(dumps(self.config))
                    pass



from json import dumps, loads
from os import path, walk
from typing import List
from jinja2 import Environment, FileSystemLoader
from marshmallow import EXCLUDE, Schema, fields
from requests import Response
from yaml import safe_load

from .http_service import HttpService
from .utils.helpers import merge_child, nested_dict_get


class AutoSwitchProviders:
    def __init__(
        self, template_dir: str, config: dict = {}, cache_config: dict = {}
    ) -> None:
        self.template_dir = template_dir
        self.providers = []
        self.config = config or {}
        self.cache_config = cache_config or None
        self.http_service = HttpService(cache_config=self.cache_config)

        if self.template_dir:
            self._discover_templates()

    def process(self, data: dict):
        if not self.providers or len(self.providers) == 0:
            print("⛔️ Empty provider!")
            return None

        process_response = None
        for provider in self.providers:
            try:
                template: dict = safe_load(provider.get("template").render(data))
            except Exception as e:
                print(f"⛔️ error - {dumps(data)}")
                print(e)
                continue

            template_type = template.get("type") or None
            template_default_config: dict = (
                self.config[template_type] if template_type else {}
            )
            template_http_service_default_config: dict = template_default_config.get(
                "http_service", {}
            )
            template_request_config = template.get("request", {})
            default_keys = template_http_service_default_config.keys()
            for default_key_item in default_keys:
                if default_key_item in template_request_config:
                    template_request_config[default_key_item] = {
                        **template_request_config.get(default_key_item),
                        **template_http_service_default_config[default_key_item],
                    }
                else:
                    template_request_config[default_key_item] = (
                        template_http_service_default_config[default_key_item]
                    )

            if "pagination" in template:
                process_response = self._paginate(
                    request=dict(
                        method=template.get("method") or "GET",
                        endpoint=template.get("url") or "/",
                        timeout=template.get("timeout", 10),
                        params=template_request_config.get("params") or None,
                        body=template_request_config.get("body") or None,
                        headers=template_request_config.get("headers") or None,
                    ),
                    config=template.get("pagination"),
                    template=template,
                )
                if process_response:
                    break
            else:
                process_response = self._single(
                    request=dict(
                        method=template.get("method") or "GET",
                        endpoint=template.get("url") or "/",
                        timeout=template.get("timeout", 10),
                        params=template_request_config.get("params") or None,
                        body=template_request_config.get("body") or None,
                        headers=template_request_config.get("headers") or None,
                    ),
                    template=template,
                )
                if process_response:
                    break

        return process_response

    def _single(self, request: dict, template: dict):
        response = self._process(request, template)
        if "response" in response and response is not None:
            return response.get("response", [])
        return None

    def _paginate(self, request: dict, config: dict, template: dict):
        next_page_type = config.get("request_field", {}).get("type")
        next_page_key = config.get("request_field", {}).get("target")

        next_cursor = None
        has_more = True

        items = []
        page = 0

        while page < config.get("total_page", 10) and has_more is True:
            page = page + 1
            if next_cursor:
                if next_page_type in request and next_page_key in request.get(
                    next_page_type
                ):
                    request[next_page_type][next_page_key] = next_cursor

            response = self._process(request, template, config)
            if "response" in response and response is not None:
                items.append(response.get("response"))
            if "has_more" in response:
                has_more = response.get("has_more")
            if "next_cursor" in response:
                next_cursor = response.get("next_cursor")

        if len(items) == 0 or not items:
            return None

        if "merge_field" not in config:
            return items

        return merge_child(items)

    def _process(self, request: dict, template: dict, pagination_config: dict = None):
        response = self.http_service.request(**request)
        cursor_response_target = (
            pagination_config.get("response_field") if pagination_config else None
        )
        next_cursor = None
        has_more = None
        process_response = []

        # Response config
        response_config: dict = template.get("response")
        if not response_config:
            return response

        response_type = response_config.get("type")
        if response_type != "text" and response:
            response_json = (
                response.json()
                if isinstance(response, Response)
                else loads(response.get("text"))
            )

            if cursor_response_target:
                next_cursor = (
                    nested_dict_get(response_json, cursor_response_target) or None
                )
                has_more = False if not next_cursor else True

            # Check errors
            if "$errors" in response_config:
                has_error: dict = self._check_response_errors(
                    response_config.get("$errors"), response
                )
                if has_error.get("status") is False:
                    return {
                        "code": 0,
                        "message": response_config.get(has_error.get("reason_field")),
                    }

            if "$schema" in response_config:
                response_json = self._transform_with_response_schema(
                    response_config.get("$schema"),
                    response_json,
                )
                process_response = response_json
            else:
                process_response = response_json

        return {
            "response": process_response,
            "has_more": has_more,
            "next_cursor": next_cursor,
        }

    @staticmethod
    def _transform_with_response_schema(schema_template: dict, response):
        schema_dict = {}
        for key in schema_template.keys():
            cur_schema_value: dict = schema_template[key]
            cur_schema_value_type: str = cur_schema_value.get("type", key)
            cur_schema_value_key: str = cur_schema_value.get("target", key)
            cur_schema_value_default: str = cur_schema_value.get("default", None)
            cur_schema_value_allow_none: bool = False

            if cur_schema_value.get("allow_none") and (
                cur_schema_value.get("allow_none") == True
                or cur_schema_value.get("allow_none") == 1
                or cur_schema_value.get("allow_none") == "True"
                or cur_schema_value.get("allow_none") == "true"
                or cur_schema_value.get("allow_none") == "1"
            ):
                cur_schema_value_allow_none = True

            if cur_schema_value_type == "list":
                list_item_schema: dict = cur_schema_value.get("$schema", {})
                item_schema_dict = {}
                for i_key in list_item_schema.keys():
                    cur_schema_item_value: dict = list_item_schema[i_key]
                    cur_schema_item_key: str = cur_schema_item_value.get("target")
                    cur_schema_item_default: str = cur_schema_item_value.get("default")
                    item_schema_dict.update(
                        {
                            i_key: fields.Raw(
                                data_key=cur_schema_item_key,
                                missing=cur_schema_item_default,
                                allow_none=cur_schema_value_allow_none,
                            )
                        }
                    )

                schema_dict.update(
                    {
                        key: fields.List(
                            cls_or_instance=fields.Nested(
                                Schema(unknown=EXCLUDE).from_dict(item_schema_dict),
                                unknown=EXCLUDE,
                            ),
                            data_key=cur_schema_value_key,
                            missing=cur_schema_value_default,
                            allow_none=cur_schema_value_allow_none,
                        )
                    }
                )
            else:
                schema_dict.update(
                    {
                        key: fields.Raw(
                            data_key=cur_schema_value_key,
                            missing=cur_schema_value_default,
                            allow_none=cur_schema_value_allow_none,
                        )
                    }
                )

        schema = Schema(unknown=EXCLUDE).from_dict(schema_dict)
        transformed_response = schema().load(response, unknown=EXCLUDE)
        return transformed_response

    @staticmethod
    def _check_response_errors(schemas: list, response):
        result = {}
        for err_schema in schemas:
            if "if" in err_schema and "then" in err_schema:
                if_cond = err_schema.get("if")
                then_cond = err_schema.get("then")
                response_status_code = (
                    response.get("status_code")
                    if not isinstance(response, Response)
                    else response.status_code
                )
                if "status_code" in if_cond and response_status_code == if_cond.get(
                    "status_code"
                ):
                    result = {
                        "status": False,
                        "reason_field": then_cond.get("reason_field", ""),
                    }
                    break

        return result

    def _discover_templates(self):
        filenames: List[str] = next(walk(self.template_dir), (None, None, []))[2]

        if len(filenames) > 0:
            jinja2_env = Environment(loader=FileSystemLoader(self.template_dir))
            self.providers = list(
                map(
                    lambda x: {
                        "_file": x,
                        "id": path.splitext(x)[0],
                        "template": jinja2_env.get_template(x),
                    },
                    filter(lambda x: x.endswith(".yaml"), filenames),
                )
            )



import sys
from os import path
import unittest

sys.path.append("./")

from src.auto_switch_providers.auto_switch_providers import AutoSwitchProviders

TEMPLATE_CONFIG = {
    "googleapi": {"http_service": {"params": {"key": ""}}},
    "proxiesapi": {"http_service": {"params": {"auth_key": ""}}},
}

CACHE_CONFIG = {"host": "127.0.0.1", "password": "", "port": 6379}


class TestSample(unittest.TestCase):
    def test_process(self):
        response = AutoSwitchProviders(
            template_dir=f"{path.dirname(__file__)}/templates",
            config=TEMPLATE_CONFIG,
        ).process({})
        self.assertEqual(response, {})


if __name__ == "__main__":
    unittest.main()



from redis import Redis, StrictRedis


class RedisClient:
    def __init__(self, config=None):
        self.config = config or {}
        if config and config.get("password"):
            self.redis_client = StrictRedis(
                host=self.config.get("host", "127.0.0.1"),
                port=self.config.get("port", 6379),
                password=self.config.get("password", ""),
                decode_responses=True,
            )
        else:
            self.redis_client = Redis(
                host=self.config.get("host", "127.0.0.1"),
                port=self.config.get("port", 6379),
                decode_responses=True,
            )

    def get_client(self):
        return self.redis_client



__title__ = "auto-switch-providers"
__description__ = "Auto Switch Providers"
__url__ = "https://github.com/fuongz/auto-switch-providers"
__version__ = "0.0.5"
__author__ = "fuongz"
__author_email__ = "hi@phuongphung.com"
__license__ = "MIT"
__copyright__ = "Copyright fuongz"



from json import dumps, loads
from urllib.parse import urlencode
import requests

from .cache_service import CacheService
from .utils.helpers import base64encode


class HttpService(object):
    def __init__(
        self,
        config=None,
        cache_config=None,
        base_url: str = None,
        default_headers=None,
        endpoints=None,
    ) -> None:
        self.config = config
        self.base_url = base_url if base_url is not None else ""
        self.default_headers = default_headers if default_headers is not None else {}
        self.endpoints = endpoints if endpoints is not None else {}
        self.cache_service = None

        if cache_config is not None:
            self.cache_service = CacheService(cache_config)

    def _request(
        self,
        endpoint: str,
        method: str = "GET",
        params: dict = None,
        body: dict = None,
        headers: dict = None,
        timeout=10,
        url_params_format=None,
    ):
        endpoint = self.endpoints.get(endpoint) or endpoint
        if url_params_format:
            endpoint = endpoint.format(**url_params_format)
        headers = (
            {**self.default_headers, **headers} if headers else self.default_headers
        )
        url = self.base_url + (endpoint or "/")
        response = requests.request(
            method=method,
            url=url,
            headers=headers,
            params=params,
            json=body,
            timeout=timeout,
        )
        return response

    def request(
        self,
        endpoint: str,
        method: str = "GET",
        params: dict = None,
        body: dict = None,
        headers: dict = None,
        timeout=10,
        status_force_list=None,
        total=3,
        url_params_format=None,
    ):
        if status_force_list is None:
            status_force_list = []

        endpoint = self.endpoints.get(endpoint) or endpoint
        if url_params_format:
            endpoint = endpoint.format(**url_params_format)

        params_str = " - " + dumps(params) if params else ""

        for _ in range(total):
            try:
                log_endpoint = self.base_url + (endpoint or "/")

                # Check exists in cache
                if self.cache_service is not None:
                    cache_params = {}

                    if params:
                        cache_params = {**cache_params, **params}
                    if body:
                        cache_params = {**cache_params, **body}
                    if headers:
                        cache_params = {**cache_params, **headers}

                    if "api_key" in cache_params:
                        cache_params.pop("api_key")
                    if "auth_key" in cache_params:
                        cache_params.pop("auth_key")

                    cache_url_query = urlencode(params or {})
                    cache_path = base64encode(cache_params)
                    found = self.cache_service.client.exists(
                        f"auto_switch_module:http_cached:{log_endpoint}:{cache_path}"
                    )
                    if found:
                        cached_response_text = self.cache_service.client.hget(
                            f"auto_switch_module:http_cached:{log_endpoint}:{cache_path}",
                            "text",
                        )
                        cached_response_status_code = self.cache_service.client.hget(
                            f"auto_switch_module:http_cached:{log_endpoint}:{cache_path}",
                            "status_code",
                        )
                        if int(cached_response_status_code) == 200:
                            print(
                                f"✅ [m] {log_endpoint}?{cache_url_query} (memory cache)"
                            )
                            return {
                                "text": loads(cached_response_text),
                                "status_code": cached_response_status_code,
                            }

                response = self._request(
                    endpoint,
                    method=method,
                    params=params,
                    body=body,
                    headers=headers,
                    timeout=timeout,
                )

                if response.status_code in status_force_list:
                    print(
                        f"✅ [{response.elapsed.total_seconds() if response else None}s] {log_endpoint}?{cache_url_query} - retrying... ({_+1}) (code: wrong status_code)"
                    )
                    continue

                if response.status_code == 200:
                    if self.cache_service is not None:
                        self.cache_service.client.hset(
                            f"auto_switch_module:http_cached:{log_endpoint}:{cache_path}",
                            "text",
                            dumps(response.text),
                        )

                        self.cache_service.client.hset(
                            f"auto_switch_module:http_cached:{log_endpoint}:{cache_path}",
                            "status_code",
                            dumps(response.status_code),
                        )

                if response.status_code == 400:
                    print(
                        f"🚸 [{timeout if timeout else None}s] {method}: {log_endpoint}?{cache_url_query} - stopped ({_+1}) (code: HTTPError) {params_str}"
                    )
                else:
                    print(
                        f"✅ [{response.elapsed.total_seconds() if response else None}s] GET: {log_endpoint}?{cache_url_query}"
                    )

                return response
            except requests.exceptions.Timeout:
                print(
                    f"⛔️ [{timeout if timeout else None}] {method}: {log_endpoint}?{cache_url_query} - retrying... ({_+1}) (code: Timeout){params_str}"
                )
                continue
            except requests.exceptions.ConnectionError:
                print(
                    f"⛔️ [{timeout if timeout else None}] {method}: {log_endpoint}?{cache_url_query} - stopped ({_+1}) (code: ConnectionError) {params_str}"
                )
                pass

        return None



import base64
import functools
from json import dumps


def nested_dict_get(dictionary, dotted_key):
    keys = dotted_key.split(".")
    return functools.reduce(
        lambda d, key: (
            d[int(key)] if isinstance(d, list) else d.get(key) if d else None
        ),
        keys,
        dictionary,
    )


def merge_child(raw_dict: dict):
    return {k: [d.get(k) for d in raw_dict] for k in set().union(*raw_dict)}


def base64encode(data):
    return base64.b64encode(dumps(data).encode("utf-8")).decode("utf-8")


def base64decode(string: str):
    return base64.b64decode(string)



import os
from pathlib import Path

from setuptools import find_packages, setup


about = {}
here = Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

with open(here / "requirements.txt") as fp:
    install_reqs = [r.rstrip() for r in fp.readlines() if not r.startswith("#")]

with open(here / "src/auto_switch_providers/__version__.py", "r") as f:
    exec(f.read(), about)


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__author_email__"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="",
    url="https://github.com/fuongz/auto-switch-providers",
    project_urls={
        "Bug Reports": "https://github.com/fuongz/auto-switch-providers/issues",
        "Source": "https://github.com/fuongz/auto-switch-providers",
    },
    license=about["__license__"],
    packages=find_packages("src", exclude=["tests"]),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.8, <4",
    install_requires=install_reqs,
)
